import sys
import numpy as np
from bblean.bitbirch import BitBirch
from bblean.fingerprints import make_fake_fingerprints, unpack_fingerprints
from inline_snapshot import snapshot


def test_random_fps_consistency() -> None:
    fps = make_fake_fingerprints(3000, n_features=2048, seed=126205095409235, pack=True)
    tree = BitBirch(branching_factor=50, threshold=0.65, merge_criterion="diameter")
    tree.fit(fps, n_features=2048)
    output_cent = tree.get_centroids()
    output_med = tree.get_medoids(fps)
    assert [c.tolist()[:5] for c in output_cent[:5]] == snapshot(
        [
            [255, 255, 255, 255, 255],
            [255, 251, 255, 255, 255],
            [255, 239, 255, 247, 255],
            [255, 255, 191, 255, 255],
            [255, 127, 235, 127, 255],
        ]
    )
    assert output_med[:5, :5].tolist() == snapshot(
        [
            [255, 255, 126, 255, 111],
            [247, 255, 255, 255, 255],
            [191, 253, 191, 255, 255],
            [255, 255, 95, 255, 239],
            [235, 255, 123, 255, 255],
        ]
    )
    # TODO For some strage reason this test *fails on macOS*
    # The kmeans implementation of sklearn seems to work different in linux and macOS
    if sys.platform != "darwin":
        tree.global_clustering(
            20,
            method="kmeans",
            n_init=1,
            init=unpack_fingerprints(np.vstack(output_cent))[::2][:20],
            max_iter=10,
        )
        output_mol_ids = tree.get_cluster_mol_ids(global_clusters=True, sort=False)
        output_med = tree.get_medoids(fps, global_clusters=True, sort=False)
        assert [o[:5] for o in output_mol_ids[:5]] == snapshot(
            [
                [16, 1023, 1793, 2, 15],
                [1873, 1882, 1912, 1954, 1970],
                [12, 1877, 1861, 2068, 2012],
                [1560, 1901, 2065, 2037, 2396],
                [62, 73, 75, 87, 121],
            ]
        )
        assert output_med[:5, :5].tolist() == snapshot(
            [
                [255, 127, 252, 111, 223],
                [255, 255, 95, 255, 239],
                [123, 239, 238, 135, 126],
                [223, 14, 207, 187, 104],
                [255, 255, 255, 247, 255],
            ]
        )



r"""SMILES manipulation"""

import typing as tp
from numpy.typing import NDArray
import numpy as np
from pathlib import Path

from bblean.utils import batched

__all__ = [
    "load_smiles",
    "calc_num_smiles",
    "iter_smiles_from_paths",
]

SmilesPaths = tp.Iterable[Path | str] | Path | str


def load_smiles(smiles_paths: SmilesPaths, max_num: int = -1) -> NDArray[np.str_]:
    r"""Simple utility to load smiles from a ``*.smi`` file"""
    smiles = []
    for i, smi in enumerate(iter_smiles_from_paths(smiles_paths)):
        if i == max_num:
            break
        smiles.append(smi)
    return np.asarray(smiles)


def calc_num_smiles(smiles_paths: SmilesPaths) -> int:
    r"""Get the total number of smiles in a sequene of paths"""
    return sum(1 for _ in iter_smiles_from_paths(smiles_paths))


def iter_smiles_from_paths(
    smiles_paths: SmilesPaths,
) -> tp.Iterator[str]:
    r"""Iterate over smiles in a sequence of smiles paths"""
    if isinstance(smiles_paths, (Path, str)):
        smiles_paths = [smiles_paths]
    for smi_path in smiles_paths:
        with open(smi_path, mode="rt", encoding="utf-8") as f:
            for smi in f:
                yield smi


def _iter_ranges_and_smiles_batches(
    smiles_paths: SmilesPaths,
    num_per_batch: int,
) -> tp.Iterable[tuple[tuple[int, int], tuple[str, ...]]]:
    start_idx = 0
    for batch in batched(iter_smiles_from_paths(smiles_paths), num_per_batch):
        size = len(batch)
        end_idx = start_idx + size
        yield (start_idx, end_idx), batch
        start_idx = end_idx


def _iter_idxs_and_smiles_batches(
    smiles_paths: SmilesPaths, num_per_batch: int
) -> tp.Iterable[tuple[int, tuple[str, ...]]]:
    yield from enumerate(batched(iter_smiles_from_paths(smiles_paths), num_per_batch))



r"""General timing tools"""

import json
from pathlib import Path
import time

from rich.console import Console


class Timer:
    def __init__(self) -> None:
        self._timings_s: dict[str, float] = {}

    @property
    def timings_s(self) -> dict[str, float]:
        return self._timings_s.copy()

    def init_timing(self, label: str = "total") -> None:
        if label in self._timings_s:
            raise ValueError(f"{label} has already been tracked")
        self._timings_s[label] = time.perf_counter()

    def end_timing(
        self, label: str = "total", console: Console | None = None, indent: bool = True
    ) -> None:
        if label not in self._timings_s:
            raise ValueError(f"{label} has not been initialized")
        self._timings_s[label] = time.perf_counter() - self._timings_s[label]
        t = self._timings_s[label]
        if console is not None:
            if indent:
                indent_str = "    "
            else:
                indent_str = ""
            if label == "total":
                console.print(f"{indent_str}- Total time elapsed: {t:.4f} s")
            else:
                console.print(f"{indent_str}- Time for {label}: {t:.4f} s")

    def dump(self, path: Path) -> None:
        with open(path, mode="wt", encoding="utf-8") as f:
            json.dump(self._timings_s, f, indent=4)



r"""Merging criteria for BitBIRCH clustering"""

import numpy as np
from numpy.typing import NDArray

# NOTE: jt_isim_from_sum is equivalent to jt_isim_diameter_compl_from_sum
from bblean.similarity import jt_isim_from_sum, jt_isim_radius_compl_from_sum

BUILTIN_MERGES = [
    "radius",
    "diameter",
    "tolerance-diameter",
    "tolerance-radius",
    "tolerance-legacy",
    "never-merge",
]


class MergeAcceptFunction:
    # For the merge functions, although outputs of jt_isim_from_sum f64, directly using
    # f64 is *not* faster than starting with uint64
    name: str = ""

    def __call__(
        self,
        threshold: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
    ) -> bool:
        raise NotImplementedError("Must be implemented by subclasses")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class RadiusMerge(MergeAcceptFunction):
    name = "radius"

    def __call__(
        self,
        threshold: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
    ) -> bool:
        return jt_isim_radius_compl_from_sum(new_ls, new_n) >= threshold


class DiameterMerge(MergeAcceptFunction):
    name = "diameter"

    def __call__(
        self,
        threshold: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
    ) -> bool:
        return jt_isim_from_sum(new_ls, new_n) >= threshold


class ToleranceDiameterMerge(MergeAcceptFunction):
    name = "tolerance-diameter"
    # NOTE: The reliability of the estimate of the cluster should be a function of the
    # size of the old cluster, so in this metric, tolerance is larger for small clusters
    # tolerance = max{ alpha * (exp(-decay * N_old) - offset), 0}

    def __init__(
        self,
        tolerance: float = 0.05,
        n_max: int = 1000,
        decay: float = 1e-3,
        adaptive: bool = True,
    ) -> None:
        self.tolerance = tolerance
        self.decay = decay
        self.offset = np.exp(-decay * n_max)
        if not adaptive:
            self.decay = 0.0
            self.offset = 0.0

    def __call__(
        self,
        threshold: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
    ) -> bool:
        new_dc = jt_isim_from_sum(new_ls, new_n)
        if new_dc < threshold:
            return False
        # If the old n is 1 then merge directly (infinite tolerance), since the
        # old_d is undefined for a single fp
        if old_n == 1:
            return True
        # Only merge if the new_dc is greater or equal to the old, up to some tolerance,
        # which decays with N
        old_dc = jt_isim_from_sum(old_ls, old_n)
        tol = max(self.tolerance * (np.exp(-self.decay * old_n) - self.offset), 0.0)
        return new_dc >= old_dc - tol

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tolerance})"


class ToleranceRadiusMerge(ToleranceDiameterMerge):
    name = "tolerance-radius"

    def __call__(
        self,
        threshold: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
    ) -> bool:
        new_rc = jt_isim_radius_compl_from_sum(new_ls, new_n)
        if new_rc < threshold:
            return False
        if old_n == 1:
            return True
        old_rc = jt_isim_radius_compl_from_sum(old_ls, old_n)
        tol = max(self.tolerance * (np.exp(-self.decay * old_n) - self.offset), 0.0)
        return new_rc >= old_rc - tol

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tolerance})"


class NeverMerge(ToleranceDiameterMerge):
    name = "never-merge"

    def __call__(
        self,
        threshold: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
    ) -> bool:
        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class ToleranceMerge(MergeAcceptFunction):
    name = "tolerance-legacy"

    def __init__(self, tolerance: float = 0.05) -> None:
        self.tolerance = tolerance

    def __call__(
        self,
        threshold: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
    ) -> bool:
        # First two branches are equivalent to 'diameter'
        new_dc = jt_isim_from_sum(new_ls, new_n)
        if new_dc < threshold:
            return False
        if old_n == 1 or nom_n != 1:
            return True
        # 'new_dc >= threshold' and 'new_n == old_n + 1' are guaranteed here
        old_dc = jt_isim_from_sum(old_ls, old_n)
        return (new_dc * new_n - old_dc * (old_n - 1)) / 2 >= old_dc - self.tolerance

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tolerance})"


def get_merge_accept_fn(
    merge_criterion: str, tolerance: float = 0.05
) -> MergeAcceptFunction:
    if merge_criterion == "radius":
        return RadiusMerge()
    elif merge_criterion == "diameter":
        return DiameterMerge()
    elif merge_criterion == "tolerance-legacy":
        return ToleranceMerge(tolerance)
    elif merge_criterion == "tolerance-diameter":
        return ToleranceDiameterMerge(tolerance)
    elif merge_criterion == "tolerance-radius":
        return ToleranceRadiusMerge(tolerance)
    elif merge_criterion == "never-merge":
        return NeverMerge(tolerance)
    raise ValueError(
        f"Unknown merge criterion {merge_criterion} "
        "Valid criteria are: radius|diameter|tolerance-diameter|tolerance-radius"
    )
